////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2016 RWS Inc, All Rights Reserved
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of version 2 of the GNU General Public License as published by
// the Free Software Foundation
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//
//*****************************************************************************
//
// PALETTE.H
//	
//*****************************************************************************

#ifndef _PALETTE_H
#define _PALETTE_H

// This is just a temporary definition of the PALETTE struct.  It may or
// may not look like this.

#define PALETTE_MAXCOLORS 127

typedef struct tag_PALETTE
{
//	RGBQUAD rgbColors[PALETTE_MAXCOLORS];
	int rgbColors[PALETTE_MAXCOLORS];
} PALETTE;

#endif //_PALETTE_H

//*****************************************************************************
// EOF
//*****************************************************************************
