//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
//#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
//LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""afxres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
#if defined(APSTUDIO_INVOKED) || defined(JAPAN_ADD_ON)
#if defined(APSTUDIO_INVOKED)
IDI_ICON3$(JAPAN_ADD_ON) ICON    DISCARDABLE     "Japan Add On.ico"
#else
IDI_ICON3               ICON    DISCARDABLE     "Japan Add On.ico"
#endif
#endif
#if defined(APSTUDIO_INVOKED) || defined(SUPER_POSTAL)
#if defined(APSTUDIO_INVOKED)
IDI_ICON1$(SUPER_POSTAL) ICON    DISCARDABLE     "Super Postal.ico"
#else
IDI_ICON1               ICON    DISCARDABLE     "Super Postal.ico"
#endif
#endif
#if defined(APSTUDIO_INVOKED) || defined(POSTAL_PLUS)
#if defined(APSTUDIO_INVOKED)
IDI_ICON2$(POSTAL_PLUS) ICON    DISCARDABLE     /*"Postal Plus.ico"*/ "guy.ico"
#else
IDI_ICON2               ICON    DISCARDABLE     /*"Postal Plus.ico"*/ "guy.ico"
#endif
#endif

/////////////////////////////////////////////////////////////////////////////
//
// Cursor
//

IDC_CURSOR1             CURSOR  DISCARDABLE     "cursor1.cur"
#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

