////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2016 RWS Inc, All Rights Reserved
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of version 2 of the GNU General Public License as published by
// the Free Software Foundation
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//
///////////////////////////////////////////////////////////////////////////////
//
//	RSPiX.h
// 
// History:
//		07/19/96 JMI	Started.
//
//////////////////////////////////////////////////////////////////////////////
//
// This file decides what platform you are on and #includes the correct
// RSPiX*.h.  If you #include the platform specific RSPiX*.h instead of this
// one, you will get a compile error.
//
//////////////////////////////////////////////////////////////////////////////
#ifndef RSPiX_H
#define RSPiX_H

#include "RSPiXUnix.h"

#endif // RSPiX_H
//////////////////////////////////////////////////////////////////////////////
// EOF
//////////////////////////////////////////////////////////////////////////////
