////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2016 RWS Inc, All Rights Reserved
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of version 2 of the GNU General Public License as published by
// the Free Software Foundation
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//
// LIST.CPP
// 
// History:
//		06/23/94 JMI	Started.
//
//		06/14/95	JMI	Converted to a template in attempt to make this class
//							more useful.  This particular file becomes unnecessary
//							at this point.  Do not bother including this CPP file
//							in your project (the entirity of this template class
//							is implemented in the .H file).
//
//
//////////////////////////////////////////////////////////////////////////////
//
// This module provides dynamic linked list services.
//
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// EOF
//////////////////////////////////////////////////////////////////////////////
